@@ -1,8 +1,10 @@
 import abc
-from typing import TypeVar, Generic, Union, Dict, List, Type
+from enum import Enum
+from typing import Dict, Generic, List, Type, TypeVar, Union
 
 import numpy as np
 import pandas as pd
+from datagen_protocol.schema.humans.human import Eyebrows, FacialHair
 
 try:
     from collections import Sequence
@@ -10,45 +12,44 @@ except ImportError:
     from collections.abc import Sequence  # >=Python 3.10
 
 from rich import pretty as pretty_rich
-from pydantic import Extra
 
-from datagen.api.catalog.attributes import AssetAttributes, Enum, AllOf, AnyOf, Exactly
-from datagen.api.assets import Human, Hair, Eyes, Glasses, Mask, Background
+from datagen.api.assets import Background, DataSequence, Eyes, Glasses, Hair, Human, Mask
+from datagen.api.catalog.attributes import AllOf, AnyOf, Exactly
 
 Asset = TypeVar("Asset")
 
 
-class AssetCreationHook(abc.ABC, Generic[Asset]):
+class AssetProvisioningHook(abc.ABC, Generic[Asset]):
     @abc.abstractmethod
     def __call__(self, *args) -> None:
         ...
 
 
-class PreProvisionHook(AssetCreationHook, abc.ABC, Generic[Asset]):
+class InitParametersHook(AssetProvisioningHook, abc.ABC, Generic[Asset]):
     @abc.abstractmethod
     def __call__(self, asset_id: str) -> dict:
         ...
 
 
-class PostLoadHook(AssetCreationHook, abc.ABC, Generic[Asset]):
+class PostLoadHook(AssetProvisioningHook, abc.ABC, Generic[Asset]):
     @abc.abstractmethod
     def __call__(self, asset: Asset) -> None:
         ...
 
 
 class AssetCreationHooks(Generic[Asset]):
-    def __init__(self, hooks: List[AssetCreationHook]):
+    def __init__(self, hooks: List[AssetProvisioningHook]):
         self._hooks = hooks
 
     @property
-    def _pre_provision_hooks(self) -> List[PreProvisionHook]:
-        return list(filter(lambda h: isinstance(h, PreProvisionHook), self._hooks))  # type: ignore
+    def _init_hooks(self) -> List[InitParametersHook]:
+        return list(filter(lambda h: isinstance(h, InitParametersHook), self._hooks))  # type: ignore
 
-    def pre_provision(self, asset_id: str) -> dict:
-        preload_params = {}
-        for h in self._pre_provision_hooks:
-            preload_params.update(**h(asset_id))
-        return preload_params
+    def get_init_params(self, asset_id: str) -> dict:
+        init_params = {}
+        for h in self._init_hooks:
+            init_params.update(**h(asset_id))
+        return init_params
 
     @property
     def _post_load_hooks(self) -> List[PostLoadHook]:
@@ -63,38 +64,27 @@ class AssetInstancesProvisioner(Generic[Asset]):
     def __init__(
         self,
         asset_type: Type[Asset],
-        asset_attributes_type: Type[AssetAttributes],
         asset_id_to_asset_attrs: dict,
-        hooks: List[AssetCreationHook[Asset]],
+        hooks: List[AssetProvisioningHook[Asset]],
     ):
         self._asset_type = asset_type
-        self._asset_attributes_type = asset_attributes_type
         self._asset_id_to_asset_attrs = asset_id_to_asset_attrs
         self._hooks = AssetCreationHooks(hooks)
 
     def provision(self, asset_id: str) -> Asset:
-        pre_provision_params = self._hooks.pre_provision(asset_id)
-        asset = self._asset_type(id=asset_id, **pre_provision_params)
-        self._add_attributes(asset)
+        asset = self._asset_type(
+            id=asset_id,
+            **self._hooks.get_init_params(asset_id),
+            attributes=self._asset_id_to_asset_attrs[asset_id],
+        )
         return asset
 
-    def _add_attributes(self, asset: Asset):
-        asset.Config.extra = Extra.allow
-        asset.attributes = self._get_attributes(asset)
-        asset.Config.extra = Extra.forbid
-
-    def _get_attributes(self, asset: Asset) -> AssetAttributes:
-        asset_attrs = self._asset_id_to_asset_attrs[asset.id]
-        if isinstance(asset_attrs, dict):
-            asset_attrs = self._asset_id_to_asset_attrs[asset.id] = self._load_attrs(asset_attrs)
-        return asset_attrs
-
-    def _load_attrs(self, attrs_dict: dict) -> AssetAttributes:
-        return self._asset_attributes_type.Schema().load(attrs_dict)
-
     def parse(self, asset_id: str, asset_body_dict: dict) -> Asset:
-        asset = self._asset_type.parse_obj({"id": asset_id, **asset_body_dict})
-        self._add_attributes(asset)
+        asset = self._asset_type(
+            id=asset_id,
+            **asset_body_dict,
+            attributes=self._asset_id_to_asset_attrs[asset_id],
+        )
         return asset
 
     def post_load(self, asset: Asset) -> Asset:
@@ -114,7 +104,7 @@ class CatalogInstancesCache(Generic[Asset]):
 
     def _load(self, asset_id) -> Asset:
         try:
-            asset = self._cache[asset_id]
+            asset = self._cache[asset_id].copy(deep=True)
         except KeyError:
             asset = self._cache.setdefault(asset_id, self._provisioner.provision(asset_id))
         return asset
@@ -174,16 +164,18 @@ class AttributesDataFrame:
         ]
         return np.logical_and.reduce(attrs_series_list)  # type: ignore
 
-    def _create_attr_series(self, attr_name: str, attr_query_val: Union[Enum, AllOf, AnyOf, Exactly]) -> pd.Series:
-        if isinstance(attr_query_val, Enum):
+    def _create_attr_series(self, attr_name: str, attr_query_val: Union[str, Enum, AllOf, AnyOf, Exactly]) -> pd.Series:
+        if isinstance(attr_query_val, (str, Enum)):
             return self._create_series(attr_name, attr_query_val)
         elif isinstance(attr_query_val, (AllOf, AnyOf)):
             series_list = [self._create_series(attr_name, attr_val) for attr_val in attr_query_val]
             operator = np.logical_and if isinstance(attr_query_val, AllOf) else np.logical_or
             return operator.reduce(series_list)  # type: ignore
 
-    def _create_series(self, attr_name: str, attr_val: Enum) -> pd.Series:
-        return self._df[f"{attr_val.value}_{attr_name}"].__eq__(True)
+    def _create_series(self, attr_name: str, attr_val: Union[str, Enum]) -> pd.Series:
+        if isinstance(attr_val, Enum):
+            attr_val = attr_val.value
+        return self._df[f"{attr_val}_{attr_name}"].__eq__(True)
 
     @staticmethod
     def from_dict(asset_id_to_asset_attrs: dict) -> "AttributesDataFrame":
@@ -200,7 +192,8 @@ class AttributesDataFrame:
             if not isinstance(attr_value, list):
                 attr_value = [attr_value]
             for v in attr_value:
-                dataframe_row[v + "_" + attr_name] = True
+                if isinstance(v, str):
+                    dataframe_row[v + "_" + attr_name] = True
         return dataframe_row
 
 
@@ -208,18 +201,14 @@ class AssetCatalog(Generic[Asset]):
     def __init__(
         self,
         asset_type: Type[Asset],
-        asset_attributes_type: Type[AssetAttributes],
         asset_id_to_asset_attrs: dict,
-        hooks: List[AssetCreationHook[Asset]] = [],
+        hooks: List[AssetProvisioningHook[Asset]] = [],
     ):
-
-        self._provisioner = AssetInstancesProvisioner(asset_type, asset_attributes_type, asset_id_to_asset_attrs, hooks)
+        self._provisioner = AssetInstancesProvisioner(asset_type, asset_id_to_asset_attrs, hooks)
         self._instances_cache = CatalogInstancesCache(self._provisioner)
         self._attributes_df = AttributesDataFrame.from_dict(asset_id_to_asset_attrs)
 
-    def get(
-        self, id: str = None, limit: int = None, **attributes
-    ) -> Union[Asset, List[Asset], Dict[str, AssetAttributes], AssetAttributes]:
+    def get(self, id: str = None, limit: int = None, **attributes) -> Union[Asset, List[Asset]]:
         if id:
             return self._instances_cache.get(id)
         else:
@@ -236,23 +225,41 @@ class AssetCatalog(Generic[Asset]):
         return self._provisioner.parse(asset_id=id, asset_body_dict=asset_body)
 
 
+class BackgroundsCatalog(AssetCatalog[Background]):
+    def get(
+        self,
+        id: str = None,
+        limit: int = None,
+        # generator=Generator.IDENTITIES,
+        **attrs,
+    ) -> Union[Asset, List[Asset]]:
+        return super().get(
+            id,
+            limit,
+            # generator=generator,
+            **attrs,
+        )
+
+
 class DatagenAssetsCatalog:
     def __init__(
         self,
         humans: AssetCatalog[Human],
+        sequences: AssetCatalog[DataSequence],
         hair: AssetCatalog[Hair],
         eyes: AssetCatalog[Eyes],
-        eyebrows: AssetCatalog[Hair],
-        beards: AssetCatalog[Hair],
+        eyebrows: AssetCatalog[Eyebrows],
+        facial_hair: AssetCatalog[FacialHair],
         glasses: AssetCatalog[Glasses],
         masks: AssetCatalog[Mask],
-        backgrounds: AssetCatalog[Background],
+        backgrounds: BackgroundsCatalog,
     ):
         self._humans = humans
+        self._sequences = sequences
         self._hair = hair
         self._eyes = eyes
         self._eyebrows = eyebrows
-        self._beards = beards
+        self._facial_hair = facial_hair
         self._glasses = glasses
         self._masks = masks
         self._backgrounds = backgrounds
@@ -263,6 +270,10 @@ class DatagenAssetsCatalog:
     def humans(self) -> AssetCatalog[Human]:
         return self._humans
 
+    @property
+    def sequences(self) -> AssetCatalog[DataSequence]:
+        return self._sequences
+
     @property
     def hair(self) -> AssetCatalog[Hair]:
         return self._hair
@@ -272,12 +283,16 @@ class DatagenAssetsCatalog:
         return self._eyes
 
     @property
-    def eyebrows(self) -> AssetCatalog[Hair]:
+    def eyebrows(self) -> AssetCatalog[Eyebrows]:
         return self._eyebrows
 
     @property
-    def beards(self) -> AssetCatalog[Hair]:
-        return self._beards
+    def facial_hair(self) -> AssetCatalog[FacialHair]:
+        return self._facial_hair
+
+    @property
+    def beards(self) -> AssetCatalog[FacialHair]:
+        return self._facial_hair
 
     @property
     def glasses(self) -> AssetCatalog[Glasses]:
@@ -288,5 +303,5 @@ class DatagenAssetsCatalog:
         return self._masks
 
     @property
-    def backgrounds(self) -> AssetCatalog[Background]:
+    def backgrounds(self) -> BackgroundsCatalog:
         return self._backgrounds
